@@ -70,6 +70,7 @@ typedef struct {
   FileOp    op;
   gpointer  stream;
   goffset   pos;
+  goffset   size;
 } FileHandle;
 
 static GThread        *subthread             = NULL;
@@ -216,6 +217,7 @@ file_handle_new (const gchar *path)
   g_mutex_init (&file_handle->mutex);
   file_handle->op = FILE_OP_NONE;
   file_handle->path = g_strdup (path);
+  file_handle->size = -1;
 
   g_hash_table_insert (global_active_fh_map, file_handle, file_handle);
 
@@ -274,6 +276,7 @@ file_handle_close_stream (FileHandle *file_handle)
       g_object_unref (file_handle->stream);
       file_handle->stream = NULL;
       file_handle->op = FILE_OP_NONE;
+      file_handle->size = -1;
     }
 }
 
@@ -734,6 +737,50 @@ file_info_get_stat_mode (GFileInfo *file_info)
   return unix_mode;
 }
 
+static void
+set_attributes_from_info (GFileInfo *file_info, struct stat *sbuf)
+{
+  GTimeVal mod_time;
+
+  sbuf->st_mode = file_info_get_stat_mode (file_info);
+  sbuf->st_size = g_file_info_get_size (file_info);
+  sbuf->st_uid = daemon_uid;
+  sbuf->st_gid = daemon_gid;
+
+  g_file_info_get_modification_time (file_info, &mod_time);
+  sbuf->st_mtime = mod_time.tv_sec;
+  sbuf->st_ctime = mod_time.tv_sec;
+  sbuf->st_atime = mod_time.tv_sec;
+
+  if (g_file_info_has_attribute (file_info, G_FILE_ATTRIBUTE_TIME_CHANGED))
+    sbuf->st_ctime = file_info_get_attribute_as_uint (file_info, G_FILE_ATTRIBUTE_TIME_CHANGED);
+  if (g_file_info_has_attribute (file_info, G_FILE_ATTRIBUTE_TIME_ACCESS))
+    sbuf->st_atime = file_info_get_attribute_as_uint (file_info, G_FILE_ATTRIBUTE_TIME_ACCESS);
+
+  if (g_file_info_has_attribute (file_info, G_FILE_ATTRIBUTE_UNIX_BLOCK_SIZE))
+    sbuf->st_blksize = file_info_get_attribute_as_uint (file_info, G_FILE_ATTRIBUTE_UNIX_BLOCK_SIZE);
+  if (g_file_info_has_attribute (file_info, G_FILE_ATTRIBUTE_UNIX_BLOCKS))
+    sbuf->st_blocks = file_info_get_attribute_as_uint (file_info, G_FILE_ATTRIBUTE_UNIX_BLOCKS);
+  else /* fake it to make 'du' work like 'du --apparent'. */
+    sbuf->st_blocks = (sbuf->st_size + 511) / 512;
+
+  /* Setting st_nlink to 1 for directories makes 'find' work */
+  sbuf->st_nlink = 1;
+}
+
+static const char *query_attributes = G_FILE_ATTRIBUTE_STANDARD_TYPE ","
+                                      G_FILE_ATTRIBUTE_STANDARD_IS_SYMLINK ","
+                                      G_FILE_ATTRIBUTE_STANDARD_SIZE ","
+                                      G_FILE_ATTRIBUTE_UNIX_MODE ","
+                                      G_FILE_ATTRIBUTE_TIME_CHANGED ","
+                                      G_FILE_ATTRIBUTE_TIME_MODIFIED ","
+                                      G_FILE_ATTRIBUTE_TIME_ACCESS ","
+                                      G_FILE_ATTRIBUTE_UNIX_BLOCK_SIZE ","
+                                      G_FILE_ATTRIBUTE_UNIX_BLOCKS ","
+                                      G_FILE_ATTRIBUTE_ACCESS_CAN_READ ","
+                                      G_FILE_ATTRIBUTE_ACCESS_CAN_WRITE ","
+                                      G_FILE_ATTRIBUTE_ACCESS_CAN_EXECUTE;
+
 static gint
 getattr_for_file (GFile *file, struct stat *sbuf)
 {
@@ -741,49 +788,11 @@ getattr_for_file (GFile *file, struct stat *sbuf)
   GError    *error  = NULL;
   gint       result = 0;
 
-  file_info = g_file_query_info (file, 
-                                 G_FILE_ATTRIBUTE_STANDARD_TYPE ","
-                                 G_FILE_ATTRIBUTE_STANDARD_NAME ","
-                                 G_FILE_ATTRIBUTE_STANDARD_IS_SYMLINK ","
-                                 G_FILE_ATTRIBUTE_STANDARD_SIZE ","
-                                 G_FILE_ATTRIBUTE_UNIX_MODE ","
-                                 G_FILE_ATTRIBUTE_TIME_CHANGED ","
-                                 G_FILE_ATTRIBUTE_TIME_MODIFIED ","
-                                 G_FILE_ATTRIBUTE_TIME_ACCESS ","
-                                 G_FILE_ATTRIBUTE_UNIX_BLOCK_SIZE ","
-                                 G_FILE_ATTRIBUTE_UNIX_BLOCKS ","
-				 "access::*",
-                                 0, NULL, &error);
+  file_info = g_file_query_info (file, query_attributes, 0, NULL, &error);
 
   if (file_info)
     {
-      GTimeVal mod_time;
-
-      sbuf->st_mode = file_info_get_stat_mode (file_info);
-      sbuf->st_size = g_file_info_get_size (file_info);
-      sbuf->st_uid = daemon_uid;
-      sbuf->st_gid = daemon_gid;
-
-      g_file_info_get_modification_time (file_info, &mod_time);
-      sbuf->st_mtime = mod_time.tv_sec;
-      sbuf->st_ctime = mod_time.tv_sec;
-      sbuf->st_atime = mod_time.tv_sec;
-
-      if (g_file_info_has_attribute (file_info, G_FILE_ATTRIBUTE_TIME_CHANGED))
-        sbuf->st_ctime = file_info_get_attribute_as_uint (file_info, G_FILE_ATTRIBUTE_TIME_CHANGED);
-      if (g_file_info_has_attribute (file_info, G_FILE_ATTRIBUTE_TIME_ACCESS))
-        sbuf->st_atime = file_info_get_attribute_as_uint (file_info, G_FILE_ATTRIBUTE_TIME_ACCESS);
-
-      if (g_file_info_has_attribute (file_info, G_FILE_ATTRIBUTE_UNIX_BLOCK_SIZE))
-	sbuf->st_blksize = file_info_get_attribute_as_uint (file_info, G_FILE_ATTRIBUTE_UNIX_BLOCK_SIZE);
-      if (g_file_info_has_attribute (file_info, G_FILE_ATTRIBUTE_UNIX_BLOCKS))
-	sbuf->st_blocks = file_info_get_attribute_as_uint (file_info, G_FILE_ATTRIBUTE_UNIX_BLOCKS);
-      else /* fake it to make 'du' work like 'du --apparent'. */
-	sbuf->st_blocks = (sbuf->st_size + 511) / 512;
-
-      /* Setting st_nlink to 1 for directories makes 'find' work */
-      sbuf->st_nlink = 1;
-
+      set_attributes_from_info (file_info, sbuf);
       g_object_unref (file_info);
     }
   else
@@ -804,16 +813,51 @@ getattr_for_file (GFile *file, struct stat *sbuf)
   return result;
 }
 
-static void
+/* Call with fh locked */
+static gint
 getattr_for_file_handle (FileHandle *fh, struct stat *sbuf)
 {
-  sbuf->st_mode = S_IFREG | S_IRUSR | S_IWUSR | S_IXUSR;
-  sbuf->st_uid = daemon_uid;
-  sbuf->st_gid = daemon_gid;
-  sbuf->st_nlink = 1;
-  sbuf->st_size = fh->pos;
-  sbuf->st_blksize = 512;
-  sbuf->st_blocks = (sbuf->st_size + 511) / 512;
+  GFileInfo *file_info;
+  GError    *error  = NULL;
+  gint       result = 0;
+
+  switch (fh->op)
+    {
+    case FILE_OP_READ:
+      file_info = g_file_input_stream_query_info (G_FILE_INPUT_STREAM (fh->stream),
+                                                  query_attributes,
+                                                  NULL, &error);
+      break;
+    case FILE_OP_WRITE:
+      file_info = g_file_output_stream_query_info (G_FILE_OUTPUT_STREAM (fh->stream),
+                                                  query_attributes,
+                                                  NULL, &error);
+      break;
+    default:
+      return -ENOTSUP;
+    }
+
+  if (file_info)
+    {
+      set_attributes_from_info (file_info, sbuf);
+      g_object_unref (file_info);
+    }
+  else
+    {
+      if (error)
+        {
+          debug_print ("Error from GVFS: %s\n", error->message);
+          result = -errno_from_error (error);
+          g_error_free (error);
+        }
+      else
+        {
+          debug_print ("No file info, but no error from GVFS.\n");
+          result = -EIO;
+        }
+    }
+
+  return result;
 }
 
 static gint
@@ -853,27 +897,50 @@ vfs_getattr (const gchar *path, struct stat *sbuf)
   else if ((file = file_from_full_path (path)))
     {
       /* Submount */
+      FileHandle *fh = get_file_handle_for_path (path);
 
-      result = getattr_for_file (file, sbuf);
-
-      if (result != 0)
+      if (fh)
         {
-          FileHandle *fh = get_file_handle_for_path (path);
+          goffset size;
 
-          /* Some backends don't create new files until their stream has
-           * been closed. So, if the path doesn't exist, but we have a stream
-           * associated with it, pretend it's there. */
+          g_mutex_lock (&fh->mutex);
 
-          if (fh != NULL)
+          result = getattr_for_file_handle (fh, sbuf);
+          size = fh->size;
+
+          g_mutex_unlock (&fh->mutex);
+          file_handle_unref (fh);
+
+          if (result == -ENOTSUP)
             {
-              g_mutex_lock (&fh->mutex);
-              getattr_for_file_handle (fh, sbuf);
-              g_mutex_unlock (&fh->mutex);
+              result = getattr_for_file (file, sbuf);
 
-              file_handle_unref (fh);
-              result = 0;
+              /* Some backends don't create new files until their stream has
+               * been closed. So, if the path doesn't exist, but we have a
+               * stream associated with it, pretend it's there. */
+
+              /* If we're tracking an open file's size, use that in preference
+               * to the stat information since it may be incorrect if
+               * g_file_replace writes to a temporary file. */
+              if (size != -1)
+                  sbuf->st_size = size;
+
+              if (result != 0)
+                {
+                  result = 0;
+                  sbuf->st_mode = S_IFREG | S_IRUSR | S_IWUSR | S_IXUSR;
+                  sbuf->st_uid = daemon_uid;
+                  sbuf->st_gid = daemon_gid;
+                  sbuf->st_nlink = 1;
+                  sbuf->st_blksize = 512;
+                  sbuf->st_blocks = (sbuf->st_size + 511) / 512;
+                }
             }
         }
+      else
+        {
+          result = getattr_for_file (file, sbuf);
+        }
 
       g_object_unref (file);
     }
@@ -921,6 +988,7 @@ setup_input_stream (GFile *file, FileHandle *fh)
           g_output_stream_close (fh->stream, NULL, NULL);
           g_object_unref (fh->stream);
           fh->stream = NULL;
+          fh->size = -1;
         }
     }
 
@@ -968,9 +1036,14 @@ setup_output_stream (GFile *file, FileHandle *fh, int flags)
   if (!fh->stream)
     {
       if (flags & O_TRUNC)
-        fh->stream = g_file_replace (file, NULL, FALSE, 0, NULL, &error);
-      else
+        {
+          fh->stream = g_file_replace (file, NULL, FALSE, 0, NULL, &error);
+          fh->size = 0;
+        }
+      else if (flags & O_APPEND)
         fh->stream = g_file_append_to (file, 0, NULL, &error);
+      else
+        result = -ENOTSUP;
       if (fh->stream)
         fh->pos = g_seekable_tell (G_SEEKABLE (fh->stream));
     }
@@ -1130,6 +1203,7 @@ vfs_create (const gchar *path, mode_t mode, struct fuse_file_info *fi)
 
               file_handle_close_stream (fh);
               fh->stream = file_output_stream;
+              fh->size = 0;
               fh->op = FILE_OP_WRITE;
 
               g_mutex_unlock (&fh->mutex);
@@ -1164,6 +1238,10 @@ vfs_release (const gchar *path, struct fuse_file_info *fi)
 
   if (fh)
     {
+      g_mutex_lock (&fh->mutex);
+      file_handle_close_stream (fh);
+      g_mutex_unlock (&fh->mutex);
+
       /* get_file_handle_from_info () adds a "working ref", so unref twice. */
       file_handle_unref (fh);
       file_handle_unref (fh);
@@ -1325,7 +1403,11 @@ vfs_read (const gchar *path, gchar *buf, size_t size,
 }
 
 static gint
-write_stream (FileHandle *fh, const gchar *input_buf, size_t input_buf_size, off_t offset)
+write_stream (FileHandle *fh,
+              gboolean is_append,
+              const gchar *input_buf,
+              size_t input_buf_size,
+              off_t offset)
 {
   GOutputStream *output_stream;
   gint           n_bytes_written = 0;
@@ -1336,7 +1418,7 @@ write_stream (FileHandle *fh, const gchar *input_buf, size_t input_buf_size, off
 
   output_stream = fh->stream;
 
-  if (offset != fh->pos)
+  if (!is_append && offset != fh->pos)
     {
       if (g_seekable_can_seek (G_SEEKABLE (output_stream)))
         {
@@ -1398,6 +1480,9 @@ write_stream (FileHandle *fh, const gchar *input_buf, size_t input_buf_size, off
         }
     }
 
+  if (fh->size != -1 && fh->pos > fh->size)
+    fh->size = fh->pos;
+
   return result;
 }
 
@@ -1418,10 +1503,11 @@ vfs_write (const gchar *path, const gchar *buf, size_t len, off_t offset,
         {
           g_mutex_lock (&fh->mutex);
 
-          result = setup_output_stream (file, fh, 0);
+          result = setup_output_stream (file, fh, fi->flags & O_APPEND);
           if (result == 0)
             {
-              result = write_stream (fh, buf, len, offset);
+              result = write_stream (fh, fi->flags & O_APPEND,
+                                     buf, len, offset);
             }
 
           g_mutex_unlock (&fh->mutex);
@@ -1447,48 +1533,6 @@ vfs_write (const gchar *path, const gchar *buf, size_t len, off_t offset,
   return result;
 }
 
-static gint
-vfs_flush (const gchar *path, struct fuse_file_info *fi)
-{
-  FileHandle *fh = get_file_handle_from_info (fi);
-
-  debug_print ("vfs_flush: %s\n", path);
-
-  if (fh)
-    {
-      g_mutex_lock (&fh->mutex);
-      file_handle_close_stream (fh);
-      g_mutex_unlock (&fh->mutex);
-
-      /* get_file_handle_from_info () adds a "working ref", so release that. */
-      file_handle_unref (fh);
-    }
-
-  /* TODO: Error handling. */
-  return 0;
-}
-
-static gint
-vfs_fsync (const gchar *path, gint sync_data_only, struct fuse_file_info *fi)
-{
-  FileHandle *fh = get_file_handle_from_info (fi);
-
-  debug_print ("vfs_flush: %s\n", path);
-
-  if (fh)
-    {
-      g_mutex_lock (&fh->mutex);
-      file_handle_close_stream (fh);
-      g_mutex_unlock (&fh->mutex);
-
-      /* get_file_handle_from_info () adds a "working ref", so release that. */
-      file_handle_unref (fh);
-    }
-
-  /* TODO: Error handling. */
-  return 0;
-}
-
 static gint
 vfs_opendir (const gchar *path, struct fuse_file_info *fi)
 {
@@ -1856,7 +1900,7 @@ pad_file (FileHandle *fh, gsize num, goffset current_size)
   buf = g_malloc0 (PAD_BLOCK_SIZE);
   for (written = 0; written < num; written += PAD_BLOCK_SIZE)
     {
-      res = write_stream (fh, buf, MIN (num - written, PAD_BLOCK_SIZE), current_size + written);
+      res = write_stream (fh, FALSE, buf, MIN (num - written, PAD_BLOCK_SIZE), current_size + written);
       if (res < 0)
         break;
     }
@@ -1865,13 +1909,73 @@ pad_file (FileHandle *fh, gsize num, goffset current_size)
   return (res < 0 ? res : 0);
 }
 
+static int
+truncate_stream (GFile *file, FileHandle *fh, off_t size)
+{
+  goffset current_size;
+  GError *error  = NULL;
+  int result = 0;
+
+  if (g_seekable_can_truncate (G_SEEKABLE (fh->stream)))
+    {
+      g_seekable_truncate (fh->stream, size, NULL, &error);
+    }
+  else if (size == 0)
+    {
+      g_output_stream_close (fh->stream, NULL, NULL);
+      g_object_unref (fh->stream);
+      fh->stream = NULL;
+
+      fh->stream = g_file_replace (file, 0, FALSE, 0, NULL, &error);
+
+      if (fh->stream != NULL)
+        {
+          /* The stream created by g_file_replace() won't always replace
+           * the file until it's been closed. So close it now to make
+           * future operations consistent. */
+          g_output_stream_close (fh->stream, NULL, NULL);
+          g_object_unref (fh->stream);
+          fh->stream = NULL;
+        }
+    }
+  else if (file_handle_get_size (fh, &current_size))
+    {
+      if (current_size == size)
+        {
+          /* Don't have to do anything to succeed */
+        }
+      else if ((current_size < size) && g_seekable_can_seek (G_SEEKABLE (fh->stream)))
+        {
+          /* If the truncated size is larger than the current size
+           * then we need to pad out the difference with 0's */
+          goffset orig_pos = g_seekable_tell (G_SEEKABLE (fh->stream));
+          result = pad_file (fh, size - current_size, current_size);
+          if (result == 0)
+            g_seekable_seek (G_SEEKABLE (fh->stream), orig_pos, G_SEEK_SET, NULL, &error);
+        }
+    }
+  else
+    {
+      result = -ENOTSUP;
+    }
+
+  if (error)
+    {
+      result = -errno_from_error (error);
+      g_error_free (error);
+    }
+
+  if (result == 0 && fh->size != -1)
+    fh->size = size;
+
+  return result;
+}
+
 static gint
 vfs_ftruncate (const gchar *path, off_t size, struct fuse_file_info *fi)
 {
   GFile  *file;
-  GError *error  = NULL;
   gint    result = 0;
-  goffset current_size;
 
   debug_print ("vfs_ftruncate: %s\n", path);
 
@@ -1885,59 +1989,10 @@ vfs_ftruncate (const gchar *path, off_t size, struct fuse_file_info *fi)
         {
           g_mutex_lock (&fh->mutex);
 
-          result = setup_output_stream (file, fh, 0);
+          result = setup_output_stream (file, fh, fi->flags & O_APPEND);
 
           if (result == 0)
-            {
-              if (g_seekable_can_truncate (G_SEEKABLE (fh->stream)))
-                {
-                  g_seekable_truncate (fh->stream, size, NULL, &error);
-                }
-              else if (size == 0)
-                {
-                  g_output_stream_close (fh->stream, NULL, NULL);
-                  g_object_unref (fh->stream);
-                  fh->stream = NULL;
-
-                  fh->stream = g_file_replace (file, 0, FALSE, 0, NULL, &error);
-
-                  if (fh->stream != NULL)
-                    {
-                      /* The stream created by g_file_replace() won't always replace
-                       * the file until it's been closed. So close it now to make
-                       * future operations consistent. */
-                      g_output_stream_close (fh->stream, NULL, NULL);
-                      g_object_unref (fh->stream);
-                      fh->stream = NULL;
-                    }
-                }
-              else if (file_handle_get_size (fh, &current_size))
-                {
-                  if (current_size == size)
-                    {
-                      /* Don't have to do anything to succeed */
-                    }
-                  else if ((current_size < size) && g_seekable_can_seek (G_SEEKABLE (fh->stream)))
-                    {
-                      /* If the truncated size is larger than the current size
-                       * then we need to pad out the difference with 0's */
-                      goffset orig_pos = g_seekable_tell (G_SEEKABLE (fh->stream));
-                      result = pad_file (fh, size - current_size, current_size);
-                      if (result == 0)
-                        g_seekable_seek (G_SEEKABLE (fh->stream), orig_pos, G_SEEK_SET, NULL, &error);
-                    }
-		}
-	      else
-		{
-		  result = -ENOTSUP;
-                }
-
-              if (error)
-                {
-                  result = -errno_from_error (error);
-                  g_error_free (error);
-                }
-            }
+            result = truncate_stream (file, fh, size);
 
           g_mutex_unlock (&fh->mutex);
           file_handle_unref (fh);
@@ -1980,27 +2035,34 @@ vfs_truncate (const gchar *path, off_t size)
       if (fh)
         g_mutex_lock (&fh->mutex);
 
-      if (size == 0)
+      if (fh->stream && fh->op == FILE_OP_WRITE)
         {
-          file_output_stream = g_file_replace (file, 0, FALSE, 0, NULL, &error);
+          result = truncate_stream (file, fh, size);
         }
       else
         {
-          file_output_stream = g_file_append_to (file, 0, NULL, &error);
+          if (size == 0)
+            {
+              file_output_stream = g_file_replace (file, 0, FALSE, 0, NULL, &error);
+            }
+          else
+            {
+              file_output_stream = g_file_append_to (file, 0, NULL, &error);
+              if (file_output_stream)
+                  g_seekable_truncate (G_SEEKABLE (file_output_stream), size, NULL, &error);
+            }
+
+          if (error)
+            {
+              result = -errno_from_error (error);
+              g_error_free (error);
+            }
+
           if (file_output_stream)
-              g_seekable_truncate (G_SEEKABLE (file_output_stream), size, NULL, &error);
-        }
-
-      if (error)
-        {
-          result = -errno_from_error (error);
-          g_error_free (error);
-        }
-
-      if (file_output_stream)
-        {
-          g_output_stream_close (G_OUTPUT_STREAM (file_output_stream), NULL, NULL);
-          g_object_unref (file_output_stream);
+            {
+              g_output_stream_close (G_OUTPUT_STREAM (file_output_stream), NULL, NULL);
+              g_object_unref (file_output_stream);
+            }
         }
 
       if (fh)
@@ -2438,8 +2500,6 @@ static struct fuse_operations vfs_oper =
   .open        = vfs_open,
   .create      = vfs_create,
   .release     = vfs_release,
-  .flush       = vfs_flush,
-  .fsync       = vfs_fsync,
 
   .read        = vfs_read,
   .write       = vfs_write,
